/*
 * file StpException.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM 
 *
 * com.ibm.rational.wvcm.stp.StpException
 *
 * (C) Copyright IBM Corporation 2006, 2010.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.rational.wvcm.stp;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Locale;

import javax.wvcm.Feedback;
import javax.wvcm.Resource;
import javax.wvcm.WvcmException;
import javax.wvcm.WvcmException.ReasonCode;

import com.ibm.rational.wvcm.stpex.StpExEnumeration;
import com.ibm.rational.wvcm.stpex.StpExEnumerationBase;

/**
 * An extension of {@link javax.wvcm.WvcmException javax.wvcm.WvcmException}
 * used throughout this API. Although API methods are declared to throw
 * WvcmException, the exception actually thrown is StpException or one of its
 * subclasses.
 * <p>
 * This class extends the WvcmException class, adding a <i>sub-reason code</i>
 * field that encodes conditions specific to this API more finely than the
 * {@link javax.wvcm.WvcmException.ReasonCode}
 * <p>
 * This class (unlike the WvcmException class) leverages the Java 1.4 chained
 * exception mechanism. By default, if an StpException is constructed from one
 * or more nested Throwables, the first nested Throwable in the list is also set
 * as the StpException's {@link java.lang.Throwable#initCause cause}. In
 * certain special cases, the cause may be set to something other than the first
 * nested Throwable.
 * 
 * Exception classes that extend StpException because they convey additional
 * information, are as follows:
 * <ul>
 * <li>{@link StpPropertyException} - adds name of property responsible for the
 * exception
 * <p>
 * <li>{@link StpPartialResultsException} - holds a ResourceList of child
 * resources successfully received, a nested list of Exceptions for child
 * resources that could not be retrieved, and an indicator of whether the two
 * lists completely identify all resources that were supposed to be retrieved.
 * <p>
 * </ul>
 */
public abstract class StpException extends WvcmException
{
    /**
     * The specification for the data implementation object associated with this
     * exception.
     */
    public interface Data {
        public String getMessage();
        public StpReasonCode getStpReasonCode();
        public String toString();
    }

    /**
     * An encoding of exception conditions specific to this API. Each
     * StpReasonCode maps to one WvcmException.ReasonCode, but, because the
     * granularity of the StpReasonCode enumeration is finer than that of the
     * ReasonCode, the mapping may be many-to-one.
     * <p>
     * 
     * When the correspondence between an StpReasonCode and a ReasonCode is not
     * obvious, the StpReasonCode is mapped to either ReasonCode.CONFLICT or 
     * ReasonCode.FORBIDDEN 
     * using the following rationale:
     * <p>
     * <ul>
     * <li>{@link javax.wvcm.WvcmException.ReasonCode#CONFLICT CONFLICT} - the exception
     * relates to an issue where the client may be able to retry the method
     * after changing the state on some object.
     * 
     * <li>{@link javax.wvcm.WvcmException.ReasonCode#FORBIDDEN FORBIDDEN} - the
     * exception relates to an issue where the client cannot do anything to
     * retry the method.
     * </ul>
     * 
     * For conditions where there is not enough information to decide whether or
     * not there is anything the client could do to retry the method, the
     * StpReasonCode is classified as CONFLICT.
     * <p>
     */

    public static enum StpReasonCode implements StpExEnumeration
    {
        /**
         * The request timed out waiting for a resource to become free.
         */
        REQUEST_TIMED_OUT(ReasonCode.ABORTED),

        /**
         * <code>Precondition:</code> Cannot create this resource at the
         * specified location.
         */
        CANNOT_CREATE_AT_THIS_LOCATION,

        /**
         * <code>Precondition:</code> Operation failed because it attempted to
         * set a protected property.
         */
        CANNOT_MODIFY_PROTECTED_PROPERTY,
        
        /** Cannot overwrite existing binding. */
        CANNOT_OVERWRITE,

        /**
         * The operation cannot be performed because of a conflict with resource
         * state.
         */
        CONFLICT,

        /**
         * A checkout was attempted with no explicit activity and no current
         * activity in the workspace.
         */
        ACTIVITY_NEEDED(ReasonCode.CONFLICT),

        /**
         * The specified record is already being edited (by the same user).
         */
        ALREADY_BEING_EDITED(ReasonCode.CONFLICT),

        /**
         * Authentication information is required but client didn't provide any.
         */
        AUTHENTICATION_INFO_REQUIRED(ReasonCode.CONFLICT),

        /**
         * The requested action is inappropriate for the current state of the
         * query (precondition failure)
         */
        BAD_SOURCE_STATE(ReasonCode.CONFLICT),

        /**
         * The checkin operation failed because the resource is not checked out.
         */
        CANNOT_CHECKIN_MUST_BE_CHECKED_OUT(ReasonCode.CONFLICT),

        /**
         * Used when a string is not supported by the server's operating system.
         */
        CANNOT_ENCODE_STRING(ReasonCode.CONFLICT),

        /**
         * The uncheckout operation failed because the resource is not checked
         * out.
         */
        CANNOT_UNCHECKOUT_MUST_BE_CHECKED_OUT(ReasonCode.CONFLICT),

        /**
         * Checkout failed because the branch was not mastered locally.
         */
        CHECKOUT_BRANCH_NOT_MASTERED(ReasonCode.CONFLICT),
        
        /**
         * Checkout failed because the branch type was not mastered locally.
         */
        CHECKOUT_BRTYPE_NOT_MASTERED(ReasonCode.CONFLICT),
        
        /** Version being checked out is not the latest  **/
        CHECKOUT_NOT_LATEST(ReasonCode.CONFLICT),

        /** The client location is not within a file area */
        CLIENT_LOCATION_NOT_IN_FILE_AREA(ReasonCode.CONFLICT),

        /** A communication precondition failed */
        CONDITIONAL_EXECUTION(ReasonCode.CONFLICT),

        /**
         * An operation failed because the connection to the remote server could
         * not be established or terminated prematurely after being established.
         */
        CONNECTION_FAILED(ReasonCode.CONFLICT),

        /**
         * An attempt to deliver a resource from the change context to the
         * database failed.
         */
        DELIVERY_ERROR(ReasonCode.CONFLICT),
        
        /** Version discordance detected  **/
        DISCORDANCE_VERSION(ReasonCode.CONFLICT),

        /** Duplicate activity name  **/
        DUPLICATE_ACTIVITY_NAME(ReasonCode.CONFLICT),

        /** Duplicate stream name  **/
        DUPLICATE_STREAM_NAME(ReasonCode.CONFLICT),

        /**
         * This code indicates that an external lock couldn't be acquired
         * because the lock already exists.
         */
        EXTERNAL_LOCK_ALREADY_PRESENT(ReasonCode.CONFLICT),

        /**
         * Some dependency required by the communication channel failed.
         */
        FAILED_DEPENDENCY(ReasonCode.CONFLICT),

        /**
         * A field did not pass validation during an attempted delivery.
         */
        FIELD_VALIDATION(ReasonCode.CONFLICT),

        /**
         * The client resource resides in a file area whose version is not
         * compatible with the currently running software.  The file area needs
         * to be upgraded in order to work with this software.
         */
        FILE_AREA_NEEDS_UPGRADE(ReasonCode.CONFLICT),

        /** A file error was encountered */
        FILE_ERROR(ReasonCode.CONFLICT),

        /**
         * While firing a named ClearQuest hook, the hook returned a message,
         * which generally represents an error or need for additional
         * information
         */
        HOOK_RETURNED_MESSAGE(ReasonCode.CONFLICT),

        /**
         * An in-line query definition during query execution
         */
        ILLEGAL_QUERY(ReasonCode.CONFLICT),

        /**
         * The client resource resides in a file area whose version is not
         * compatible with the currently running software. The software needs to
         * be upgraded to handle this file area.
         */
        INCOMPATIBLE_FILE_AREA_VERSION(ReasonCode.CONFLICT),

        /** Insufficient permission for the requested operation */
        INSUFFICIENT_PERMISSION(ReasonCode.CONFLICT),

        /** An interaction request has occurred */
        INTERACTION_REQUEST(ReasonCode.CONFLICT),

        /** An internal error has occurred */
        INTERNAL_ERROR(ReasonCode.CONFLICT),

        /**
         * The provider's server encountered an unexpected internal error
         * condition which prevented it from fulfilling the request. This
         * loosely corresponds to an HTTP 500 Internal Server Error response
         * from the server.
         * 
         * @see #SERVER_ERROR
         */
        INTERNAL_SERVER_ERROR(ReasonCode.CONFLICT),

        /**
         * A required field is missing from or malformed in an StpLocation
         * specification.
         */
        INVALID_OBJECT_SELECTOR(ReasonCode.CONFLICT),

        /** An invalid response was received */
        INVALID_RESPONSE(ReasonCode.CONFLICT),

        /**
         * Completion of operation was prevented because one or more properties
         * have invalid values
         */
        INVALID_VALUES(ReasonCode.CONFLICT),
        
        /** License error occurred **/
        LICENSE_ERROR(ReasonCode.CONFLICT),

        /** The database to be affected is currently locked */
        LOCKED_DATABASE(ReasonCode.CONFLICT),

        /** The name supplied for a new resource is invalid */
        NAME_MUST_BE_VALID(ReasonCode.CONFLICT),

        /** Resource needs to be merged from latest version  **/
        NEEDS_MERGE_FROM_LATEST(ReasonCode.CONFLICT),

        /** For use when StpException is just a wrapper for a WvcmException */
        NONE(ReasonCode.CONFLICT),

        /**
         * A duplicate record is specified but the action is not a duplicate
         * action.
         */
        NOT_DUPLICATE_ACTION(ReasonCode.CONFLICT),

        /** A parameter mismatch was detected in a response */
        PARAMETER_MISMATCH(ReasonCode.CONFLICT),

        /** The parent of a targeted resource needs to exist, but doesn't */
        PARENT_MUST_EXIST(ReasonCode.CONFLICT),

        /**
         * This exception is reporting failure in an operation that was applied
         * independently to multiple resources and failed on some of them.
         * 
         * @see StpPartialResultsException
         */
        PARTIAL_RESULTS(ReasonCode.CONFLICT),

        /** Used when delivering change contexts */
        PRIOR_COMMIT_FAILURE(ReasonCode.CONFLICT),

        /**
         * Some other property error such as
         * <ul>
         * <li>Can't update value because it is inappropriate for property.
         * <li>Can't update value because of server-specific restriction such
         * as length of string or list.
         * <li>
         * </ul>
         */
        PROPERTY_ERROR(ReasonCode.CONFLICT),

        /**
         * An exception with this StpReasonCode is thrown by the execution of
         * any property "getter" method when the targeted property could not be
         * retrieved from the server. Exceptions of this type wrap the exception
         * generated by the server, which is accessible via the getCause() or
         * {@link javax.wvcm.WvcmException#getNestedExceptions()} methods of
         * this wrapping exception.
         * <p>
         * The traceback for the outer, PROPERTY_RETRIEVAL_FAILED exception will
         * identify the context in which the attempt was made to get the
         * property value from the proxy, while the traceback for the cause of
         * that exception will identify the context in which the attempt was
         * made to read the value into the proxy.
         */
        PROPERTY_RETRIEVAL_FAILED(ReasonCode.CONFLICT),

        /**
         * Thrown by CreateRecord when an attempt is made to create a record with
         * the same name as one that already exists on the server.
         */
        RECORD_WITH_SAME_DISPLAYNAME_EXISTS(ReasonCode.CONFLICT),

        /** Request failed error **/
        REQUEST_FAILED_ERROR(ReasonCode.CONFLICT),

        /**
         * The provider has detected something inappropriate with a server's
         * response. It could be the result of a bug in the server's response or
         * a bug in the provider's processing of the response.
         * 
         * @see #INTERNAL_SERVER_ERROR
         */
        SERVER_ERROR(ReasonCode.CONFLICT),

        /** View update cancel failed **/
        SYNC_CANCEL_FAILED(ReasonCode.CONFLICT),

        /** 
         * View's config spec is not synchronized with stream's configuration.
         * An update view operation is required.
         */
        VIEW_OUT_OF_SYNC_WITH_STREAM(ReasonCode.CONFLICT),

        /**
         * The provider was unable to complete the operation for an unspecified
         * reason.
         */
        FORBIDDEN,

        /** Request not understood or contextually incorrect for the provider. */
        BAD_REQUEST(ReasonCode.FORBIDDEN),

        /** Used by REVERT method */
        CHILD_ORIGINAL_SOURCE_DIRECTORY_NO_LONGER_EXISTS(ReasonCode.FORBIDDEN),

        /** Used by REVERT method */
        CHILDREN_OF_FOLDER_MUST_BE_REVERTED_FIRST(ReasonCode.FORBIDDEN),

        /**
         * Used when an operation is forbidden due to operating in disconnected
         * mode
         */
        DISCONNECTED(ReasonCode.FORBIDDEN),

        /** Used when trying to delete query folders */
        FOLDER_HAS_CHILDREN(ReasonCode.FORBIDDEN),

        /** An illegal argument was specified */
        ILLEGAL_ARG(ReasonCode.FORBIDDEN),

        /** The request or operation in not valid */
        INVALID(ReasonCode.FORBIDDEN),

        /**
         * Thrown by OpenRecord when a duplicate record is not specified with a
         * duplicate action.
         */
        NO_DUPLICATE_RECORD(ReasonCode.FORBIDDEN),

        /** Request not allowed by the provider. */
        NOT_ALLOWED(ReasonCode.FORBIDDEN),

        /** The request or operation is not supported */
        NOT_SUPPORTED(ReasonCode.FORBIDDEN),

        /** The submit request is not allowed */
        SUBMIT_NOT_ALLOWED(ReasonCode.FORBIDDEN),

        /**
         * Thrown by OpenRecord when the specified action is not defined for the
         * record type.
         */
        UNKNOWN_ACTION(ReasonCode.FORBIDDEN),

        /**
         * Thrown when a report has been requested on resource that does not
         * support that report type.
         */
        UNSUPPORTED_REPORT(ReasonCode.FORBIDDEN),

        /** Illegal syntax for location string value. */
        ILLEGAL_LOCATION_SYNTAX,

        /**
         * <code>Precondition:</code> Report failed since the resource does
         * not support the specified report.
         */
        BAD_REPORT(ReasonCode.METHOD_NOT_SUPPORTED),

        /** The resource has no content. */
        NO_CONTENT(ReasonCode.METHOD_NOT_SUPPORTED),

        /**
         * The corresponding remote resource no longer exists or was never
         * created.
         */
        NOT_FOUND,
        
        /**
         * <code>Precondition:</code> Failed to retrieve a property that
         * should be supported. A potentially recoverable condition prevented
         * the server from retrieving the property value.
         */
        PROPERTY_NOT_CURRENTLY_AVAILABLE,       
        
        /**
         * The requested property value is unavailable because it is not valid
         * for the targeted resource--the property name used is not defined in
         * the targeted resource's interface nor is it included in the
         * PropertyRequest returned by
         * {@link javax.wvcm.Resource#doGetPropertyNameList(Feedback)} for the
         * resource.
         */
        PROPERTY_NOT_DEFINED_FOR_RESOURCE,
        
        /**
         * The property value is maintained only on the server and so is not
         * available locally
         */
        PROPERTY_NOT_AVAILABLE_LOCALLY(ReasonCode.PROPERTY_NOT_DEFINED_FOR_RESOURCE),

        /**
         * The property value is unavailable because it was not in the property
         * name list when the proxy was created nor has it subsequently been
         * added via {@link Resource#setProperty} or one of the other property
         * setters.
         */
        PROPERTY_NOT_REQUESTED,

        /**
         * Even though this API says the property is valid for the targeted
         * resource, the server does not support it. For properties the server
         * intends to support in the release under development, the exception
         * message should say NOT YET IMPLEMENTED.
         */
        PROPERTY_NOT_SUPPORTED_BY_SERVER,

        /** The property value update would overwrite an earlier change. */
        PROPERTY_OVERWRITE_FORBIDDEN,
        
        /** The provider suffered an I/O failure, the operation may be retried. */
        READ_FAILED,

        /**
         * <code>Precondition:</code> Creating a resource failed because a
         * resource already exists at the specified location.
         */
        RESOURCE_ALREADY_EXISTS_AT_LOCATION,
        
        /** View update was canceled **/
        SYNC_CANCELLED (ReasonCode.CONFLICT),

        /** The user is not authorized to execute the attempted operation. */
        UNAUTHORIZED,

        /** Login on server failed**/
        LOGIN_FAILED(ReasonCode.UNAUTHORIZED),
        
        /** The provider suffered an I/O failure, the operation may be retried. */
        WRITE_FAILED,
        
        /**
         * Session does not exist or the session has expired.
         */
        SESSION_EXPIRED_OR_DOES_NOT_EXIST(ReasonCode.UNAUTHORIZED),
        
        /**
         * Used if an operation requires credentials but none were provided
         */
        CREDENTIALS_REQUIRED(ReasonCode.UNAUTHORIZED),
        
        /**
         * Used when the server capacity has been reached.
         */
        SERVER_BUSY(ReasonCode.CONFLICT),
        
        /**
         * Used when the client is not compatible with the server.  (This may
         * be caused by a client that is too old, or too new.)
         */
        INCOMPATIBLE_SERVER(ReasonCode.VERSION_NOT_SUPPORTED),
        
        RPC_UNEXPECTEDLY_EXITED(ReasonCode.CONFLICT),
        
        ;
        /* end of enum StpReasonCode */;
        
        // ====================================================================

        /**
         * Returns the WVCM base reason code associated with this STP reason
         * code.
         * 
         * @return The WVCM reason code associated with this STP reason code.
         */
        public ReasonCode getWvcmReasonCode()
        {
            return m_wvcmReasonCode;
        }

        /*
         * A string representation of this reason code.
         */
        public String toString()
        {
            return name().toLowerCase().replaceAll("_", "-");
        }

        /**
         * Constructs an StpReasonCode that is semantically equivalent to the
         * WVCM ReasonCode of the same name.
         */
        private StpReasonCode()
        {
            m_wvcmReasonCode = ReasonCode.valueOf(ReasonCode.class, name());
        }

        /**
         * Constructs an StpReasonCode as a refinement of a WVCM base
         * ReasonCode.
         * 
         * @param wvcmBaseCode The WVCM reason code that this StpReasonCode
         *            refines. Cannot be <code>null</code>.
         */
        private StpReasonCode(ReasonCode wvcmBaseCode)
        {
            m_wvcmReasonCode = wvcmBaseCode;
        }

        /**
         * The WVCM reason code classification for this SubReasonCode.
         */
        private final ReasonCode m_wvcmReasonCode;
    }

    /**
     * Casts an Object to a Type, avoiding a type safety warning. Use sparingly.
     * May throw ClassCastException at runtime. Some Java compilers can deduce
     * U from context, such as in an assignment or a return statement; however,
     * others may not. It is suggested, therefore, that all uses of this method
     * should include the target type explicitly.
     * <pre>
     *  StpException.&lt;desired-type>unchecked_cast(x)
     * </pre> 
     * The ugliness of this construct matches the sledge hammer that is being
     * used to make the code compile without warnings.
     * 
     * @param <U> The Type to which the object should be cast
     * @param obj The Object to be cast
     * @return The argument Object cast to Type U.
     */
    @SuppressWarnings("unchecked")
    public static <U> U unchecked_cast(Object obj)
    {
        return (U)obj;
    }
    
    /**
     * @return Returns an implementation object that stores the fields of this
     *         exception not defined by WvcmException and implements the other
     *         methods of this exception. Will never be <b>null</b>.
     */
    public abstract Data data();

    /**
     * Localizes the message contained within this exception and returns it.
     */
    public String getMessage() { return data().getMessage(); }
    
    /**
     * @return The StpReasonCode assigned to this exception.
     */
    public StpReasonCode getStpReasonCode()
    { return data().getStpReasonCode(); }
    
    /**
     * @return A String image of this StpException and any nested Exceptions
     */
    public String toString() { return data().toString(); }

    /**
     * Constructs this exception object for its subclasses.
     * 
     * @param resource The Resource argument to WvcmException
     * @param reasonCode The ReasonCode argument to WvcmException
     * @param nestedExceptions The Throwable[] argument to WvcmException
     * 
     * @see javax.wvcm.WvcmException
     */
    protected StpException(Resource resource,
                           ReasonCode reasonCode,
                           Throwable... nestedExceptions)
    {
        super(null, resource, reasonCode, nestedExceptions);
    }
}
